@@ -1,14 +1,17 @@
 /*********************************************************************
- *   Copyright 2009, UCAR/Unidata
+ *   Copyright 2018, UCAR/Unidata
  *   See netcdf/COPYRIGHT file for copying and redistribution conditions.
  *********************************************************************/
 /* $Id: data.c,v 1.7 2010/05/24 19:59:56 dmh Exp $ */
 /* $Header: /upc/share/CVS/netcdf-3/ncgen/data.c,v 1.7 2010/05/24 19:59:56 dmh Exp $ */
 
 #include        "includes.h"
-#include        "offsets.h"
+#include        "ncoffsets.h"
+#include        "netcdf_aux.h"
 #include        "dump.h"
 
+#undef VERIFY
+
 #define XVSNPRINTF vsnprintf
 /*
 #define XVSNPRINTF lvsnprintf
@@ -18,12 +21,12 @@ extern int lvsnprintf(char*, size_t, const char*, va_list);
 #define DATALISTINIT 32
 
 /* Track all known datalist*/
-Datalist* alldatalists = NULL;
+List* alldatalists = NULL;
 
 NCConstant nullconstant;
 NCConstant fillconstant;
 
-Datalist nildatalist; /* to support NIL keyword */
+Datalist* filldatalist;
 
 Bytebuffer* codebuffer;
 Bytebuffer* codetmp;
@@ -31,32 +34,31 @@ Bytebuffer* stmt;
 
 
 /* Forward */
+static void setconstlist(NCConstant* con, Datalist* dl);
 
-/**************************************************/
-/**************************************************/
-
-/* return 1 if the next element in the datasrc is compound*/
-int
-issublist(Datasrc* datasrc) {return istype(datasrc,NC_COMPOUND);}
-
-/* return 1 if the next element in the datasrc is a string*/
-int
-isstring(Datasrc* datasrc) {return istype(datasrc,NC_STRING);}
-
-/* return 1 if the next element in the datasrc is a fill value*/
-int
-isfillvalue(Datasrc* datasrc)
+#ifdef VERIFY
+/* index of match */
+static int
+verify(List* all, Datalist* dl)
 {
-return srcpeek(datasrc) == NULL || istype(datasrc,NC_FILLVALUE);
+    int i;
+    for(i=0;i<listlength(all);i++) {
+	void* pi = listget(all,i);
+	if(pi == dl)
+	    return i;
+    }
+    return -1;
 }
+#endif
 
-/* return 1 if the next element in the datasrc is nc_type*/
-int
-istype(Datasrc* datasrc , nc_type nctype)
+/**************************************************/
+/**************************************************/
+
+NCConstant*
+nullconst(void)
 {
-    NCConstant* ci = srcpeek(datasrc);
-    if(ci != NULL && ci->nctype == nctype) return 1;
-    return 0;
+    NCConstant* n = ecalloc(sizeof(NCConstant));
+    return n;
 }
 
 int
@@ -72,61 +74,15 @@ isstringable(nc_type nctype)
     return 0;
 }
 
-/**************************************************/
-
-void
-freedatasrc(Datasrc* src)
-{
-    efree(src);
-}
-
-Datasrc*
-allocdatasrc(void)
-{
-    Datasrc* src;
-    src = emalloc(sizeof(Datasrc));
-    src->data = NULL;
-    src->index = 0;
-    src->length = 0;
-    src->prev = NULL;
-    return src;
-}
-
-Datasrc*
-datalist2src(Datalist* list)
-{
-    Datasrc* src;
-    ASSERT(list != NULL);
-    src = allocdatasrc();
-    src->data = list->data;
-    src->index = 0;
-    src->length = list->length;
-    DUMPSRC(src,"#");
-    return src;
-}
-
-Datasrc*
-const2src(NCConstant* con)
-{
-    Datasrc* src;
-    ASSERT(con != NULL);
-    src = allocdatasrc();
-    src->data = con;
-    src->index = 0;
-    src->length = 1;
-    DUMPSRC(src,"#");
-    return src;
-}
-
-NCConstant
+NCConstant*
 list2const(Datalist* list)
 {
-    NCConstant con;
+    NCConstant* con = nullconst();
     ASSERT(list != NULL);
-    con.nctype = NC_COMPOUND;
-    con.lineno = list->data[0].lineno;
-    con.value.compoundv = list;
-    con.filled = 0;
+    con->nctype = NC_COMPOUND;
+    if(!list->readonly) con->lineno = list->data[0]->lineno;
+    setconstlist(con,list);
+    con->filled = 0;
     return con;
 }
 
@@ -142,109 +98,6 @@ const2list(NCConstant* con)
     return list;
 }
 
-NCConstant*
-srcpeek(Datasrc* ds)
-{
-    if(ds == NULL) return NULL;
-    if(ds->index < ds->length)
-	return &ds->data[ds->index];
-    if(ds->spliced)
-	return srcpeek(ds->prev);
-    return NULL;
-}
-
-void
-srcreset(Datasrc* ds)
-{
-    ds->index = 0;
-}
-
-NCConstant*
-srcnext(Datasrc* ds)
-{
-    DUMPSRC(ds,"!");
-    if(ds == NULL) return NULL;
-    if(ds->index < ds->length)
-	return &ds->data[ds->index++];
-    if(ds->spliced) {
-	srcpop(ds);
-	return srcnext(ds);
-    }
-    return NULL;
-}
-
-int
-srcmore(Datasrc* ds)
-{
-    if(ds == NULL) return 0;
-    if(ds->index < ds->length) return 1;
-    if(ds->spliced) return srcmore(ds->prev);
-    return 0;
-}
-
-int
-srcline(Datasrc* ds)
-{
-    int index = ds->index;
-    int len = ds->length;
-    /* pick closest available entry*/
-    if(len == 0) return 0;
-    if(index >= len) index = len-1;
-    return ds->data[index].lineno;
-}
-
-void
-srcpush(Datasrc* src)
-{
-    NCConstant* con;
-    ASSERT(src != NULL);
-    con = srcnext(src);
-    ASSERT(con->nctype == NC_COMPOUND);
-    srcpushlist(src,con->value.compoundv);
-}
-
-void
-srcpushlist(Datasrc* src, Datalist* dl)
-{
-    Datasrc* newsrc;
-    ASSERT(src != NULL && dl != NULL);
-    newsrc = allocdatasrc();
-    *newsrc = *src;
-    src->prev = newsrc;
-    src->index = 0;
-    src->data = dl->data;
-    src->length = dl->length;
-    DUMPSRC(src,">!");
-}
-
-void
-srcpop(Datasrc* src)
-{
-    if(src != NULL) {
-        Datasrc* prev = src->prev;
-	*src = *prev;
-        freedatasrc(prev);
-    }
-    DUMPSRC(src,"<");
-}
-
-void
-srcsplice(Datasrc* ds, Datalist* list)
-{
-    srcpushlist(ds,list);
-    ds->spliced = 1;    
-}
-
-void
-srcsetfill(Datasrc* ds, Datalist* list)
-{
-    if(ds->index >= ds->length) PANIC("srcsetfill: no space");
-    if(ds->data[ds->index].nctype != NC_FILLVALUE) PANIC("srcsetfill: not fill");
-    ds->data[ds->index].nctype = NC_COMPOUND;
-    ds->data[ds->index].value.compoundv = list;
-}
-
-
 /**************************************************/
 #ifdef GENDEBUG
 void
@@ -258,93 +111,98 @@ fflush(stderr);
 bbFree(buf);
 }
 
-void
-report0(char* lead, Datasrc* src, int index)
-{
-}
 #endif
 
 /**************************************************/
 
-/* Shallow constant cloning*/
-NCConstant
+static void
+setconstlist(NCConstant* con, Datalist* dl)
+{
+#ifdef VERIFY
+    int pos = verify(alldatalists,dl);
+    if(pos >= 0) {
+        dumpdatalist(listget(alldatalists,pos),"XXX");
+    }
+#endif
+    con->value.compoundv = dl;
+}
+
+
+/* Deep constant cloning; return struct not pointer to struct*/
+NCConstant*
 cloneconstant(NCConstant* con)
 {
-    NCConstant newcon = *con;
-    char* s;
-    switch (newcon.nctype) {
+    NCConstant* newcon = NULL;
+    Datalist* newdl = NULL;
+    char* s = NULL;
+
+    newcon = nullconst();
+    if(newcon == NULL) return newcon;
+    *newcon = *con;
+    switch (newcon->nctype) {
     case NC_STRING:
-	s = (char*)emalloc(newcon.value.stringv.len+1);
-	memcpy(s,newcon.value.stringv.stringv,newcon.value.stringv.len);
-	s[newcon.value.stringv.len] = '\0';
-	newcon.value.stringv.stringv = s;
+	if(newcon->value.stringv.len == 0)
+	    s = NULL;
+	else {
+	    s = (char*)ecalloc(newcon->value.stringv.len+1);
+	    if(newcon->value.stringv.len > 0)
+	        memcpy(s,newcon->value.stringv.stringv,newcon->value.stringv.len);
+	    s[newcon->value.stringv.len] = '\0';
+	}
+	newcon->value.stringv.stringv = s;
 	break;
     case NC_OPAQUE:
-	s = (char*)emalloc(newcon.value.opaquev.len+1);
-	memcpy(s,newcon.value.opaquev.stringv,newcon.value.opaquev.len);
-	s[newcon.value.opaquev.len] = '\0';
-	newcon.value.opaquev.stringv = s;
+	s = (char*)ecalloc(newcon->value.opaquev.len+1);
+	if(newcon->value.opaquev.len > 0)
+	    memcpy(s,newcon->value.opaquev.stringv,newcon->value.opaquev.len);
+	s[newcon->value.opaquev.len] = '\0';
+	newcon->value.opaquev.stringv = s;
+	break;
+    case NC_COMPOUND:
+	newdl = clonedatalist(con->value.compoundv);
+	setconstlist(newcon,newdl);
 	break;
     default: break;
     }
     return newcon;
 }
 
-/**************************************************/
-
-Datalist*
-datalistclone(Datalist* dl)
+/* Deep constant clear*/
+void
+clearconstant(NCConstant* con)
 {
-    int i;
-    Datalist* clone = builddatalist(dl->length);
-    for(i=0;i<dl->length;i++) {
-	clone->data[i] = cloneconstant(dl->data+i);
+    if(con == NULL) return;
+    switch (con->nctype) {
+    case NC_STRING:
+	if(con->value.stringv.stringv != NULL)
+	    efree(con->value.stringv.stringv);
+	break;
+    case NC_OPAQUE:
+	if(con->value.opaquev.stringv != NULL)
+	    efree(con->value.opaquev.stringv);
+	break;
+    case NC_COMPOUND:
+	con->value.compoundv = NULL;
+	break;
+    default: break;
     }
-    return clone;
+    memset((void*)con,0,sizeof(NCConstant));
 }
 
-Datalist*
-datalistconcat(Datalist* dl1, Datalist* dl2)
+void
+freeconstant(NCConstant* con, int shallow)
 {
-    NCConstant* vector;
-    ASSERT(dl1 != NULL);
-    if(dl2 == NULL) return dl1;
-    vector = (NCConstant*)erealloc(dl1->data,sizeof(NCConstant)*(dl1->length+dl2->length));
-    if(vector == NULL) return NULL;
-    memcpy((void*)(vector+dl1->length),dl2->data,sizeof(NCConstant)*(dl2->length));
-    dl1->data = vector;
-    return dl1;
+    if(!shallow) clearconstant(con);
+    nullfree(con);
 }
 
-Datalist*
-datalistappend(Datalist* dl, NCConstant* con)
-{
-    NCConstant* vector;
-    ASSERT(dl != NULL);
-    if(con == NULL) return dl;
-    vector = (NCConstant*)erealloc(dl->data,sizeof(NCConstant)*(dl->length+1));
-    if(vector == NULL) return NULL;
-    vector[dl->length] = *con;
-    dl->length++;
-    dl->data = vector;
-    return dl;
-}
-
-Datalist*
-datalistreplace(Datalist* dl, unsigned int index, NCConstant* con)
-{
-    ASSERT(dl != NULL);
-    ASSERT(index < dl->length);
-    ASSERT(con != NULL);
-    dl->data[index] = *con;
-    return dl;
-}
+/**************************************************/
 
 int
 datalistline(Datalist* ds)
 {
     if(ds == NULL || ds->length == 0) return 0;
-    return ds->data[0].lineno;
+    return ds->data[0]->lineno;
 }
 
 
@@ -441,19 +299,22 @@ wordstring(char* p, Bytebuffer* buf, int quote)
 
 static const char zeros[] =
     "\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0";
+
 void
 alignbuffer(NCConstant* prim, Bytebuffer* buf)
 {
     int alignment,pad,offset;
 
+    ASSERT(prim->nctype != NC_COMPOUND);
+
     if(prim->nctype == NC_ECONST)
-        alignment = nctypealignment(prim->value.enumv->typ.typecode);
+        alignment = ncaux_class_alignment(prim->value.enumv->typ.typecode);
     else if(usingclassic && prim->nctype == NC_STRING)
-        alignment = nctypealignment(NC_CHAR);
+        alignment = ncaux_class_alignment(NC_CHAR);
     else if(prim->nctype == NC_CHAR)
-        alignment = nctypealignment(NC_CHAR);
+        alignment = ncaux_class_alignment(NC_CHAR);
     else
-        alignment = nctypealignment(prim->nctype);
+        alignment = ncaux_class_alignment(prim->nctype);
     offset = bbLength(buf);
     pad = getpadding(offset,alignment);
     if(pad > 0) {
@@ -553,10 +414,10 @@ retry:	    switch ((c=*p++)) {
 	    case 'u':
 		if(hcount == 2) {
    	            snprintf(tmp,sizeof(tmp),"%hhu",
-			(unsigned int)va_arg(argv,unsigned int));
+			(unsigned char)va_arg(argv,unsigned int));
 		} else if(hcount == 1) {
    	            snprintf(tmp,sizeof(tmp),"%hu",
-			(unsigned int)va_arg(argv,unsigned int));
+			(unsigned short)va_arg(argv,unsigned int));
 		} else if(lcount == 2) {
    	            snprintf(tmp,sizeof(tmp),"%llu",
 			(unsigned long long)va_arg(argv,unsigned long long));
@@ -572,10 +433,10 @@ retry:	    switch ((c=*p++)) {
 	    case 'd':
 		if(hcount == 2) {
    	            snprintf(tmp,sizeof(tmp),"%hhd",
-			(signed int)va_arg(argv,signed int));
+			(signed char)va_arg(argv,signed int));
 		} else if(hcount == 1) {
    	            snprintf(tmp,sizeof(tmp),"%hd",
-			(signed int)va_arg(argv,signed int));
+			(signed short)va_arg(argv,signed int));
 		} else if(lcount == 2) {
    	            snprintf(tmp,sizeof(tmp),"%lld",
 			(signed long long)va_arg(argv,signed long long));
@@ -645,19 +506,21 @@ codeprintf(const char *fmt, ...)
 }
 
 NCConstant*
-emptycompoundconst(int lineno, NCConstant* c)
+emptycompoundconst(int lineno)
 {
-    ASSERT(c != NULL);
+    NCConstant* c = nullconst();
     c->lineno = lineno;
     c->nctype = NC_COMPOUND;
-    c->value.compoundv = builddatalist(0);
+    setconstlist(c,builddatalist(0));
     c->filled = 0;
     return c;    
 }
 
+/* Make an empty string constant*/
 NCConstant*
-emptystringconst(int lineno, NCConstant* c)
+emptystringconst(int lineno)
 {
+    NCConstant* c = nullconst();
     ASSERT(c != NULL);
     c->lineno = lineno;
     c->nctype = NC_STRING;
@@ -675,7 +538,7 @@ indented(int n)
 {
     char* indentation;
     if(dent == NULL) {
-	dent = (char*)emalloc(INDENTMAX+1);
+	dent = (char*)ecalloc(INDENTMAX+1);
 	memset((void*)dent,' ',INDENTMAX);
 	dent[INDENTMAX] = '\0';	
     }
@@ -685,39 +548,46 @@ indented(int n)
 }
 
 void
-dlextend(Datalist* dl)
+dlsetalloc(Datalist* dl, size_t need)
 {
-    size_t newalloc;
-    newalloc = (dl->alloc > 0?2*dl->alloc:1);
-    dlsetalloc(dl,newalloc);
+    NCConstant** newdata = NULL;
+    if(dl->readonly) abort();
+    if(dl->alloc < need) {
+        newdata = (NCConstant**)ecalloc(need*sizeof(NCConstant*));
+        if(dl->length > 0)
+            memcpy(newdata,dl->data,sizeof(NCConstant*)*dl->length);
+        dl->alloc = need;
+        nullfree(dl->data);
+        dl->data = newdata;
+    }
 }
 
 void
-dlsetalloc(Datalist* dl, size_t newalloc)
+dlextend(Datalist* dl)
 {
-    NCConstant* newdata;
-    if(newalloc <= 0) newalloc = 1;
-    if(dl->alloc > 0)
-        newdata = (NCConstant*)erealloc((void*)dl->data,sizeof(NCConstant)*newalloc);
-    else {
-        newdata = (NCConstant*)emalloc(sizeof(NCConstant)*newalloc);
-        memset((void*)newdata,0,sizeof(NCConstant)*newalloc);
-    }
-    dl->alloc = newalloc;
-    dl->data = newdata;
+    size_t newalloc;
+    if(dl->readonly) abort();
+    newalloc = (dl->alloc > 0?2*dl->alloc:2);
+    dlsetalloc(dl,newalloc);
 }
 
 
+void
+capture(Datalist* dl)
+{
+    if(alldatalists == NULL) alldatalists = listnew();
+    listpush(alldatalists,dl);
+}
+
 Datalist*
 builddatalist(int initial)
 {
     Datalist* ci;
     if(initial <= 0) initial = DATALISTINIT;
     initial++; /* for header*/
-    ci = (Datalist*)emalloc(sizeof(Datalist));
-    memset((void*)ci,0,sizeof(Datalist)); /* only clear the hdr*/
-    ci->data = (NCConstant*)emalloc(sizeof(NCConstant)*initial);
-    memset((void*)ci->data,0,sizeof(NCConstant)*initial);
+    ci = (Datalist*)ecalloc(sizeof(Datalist));
+    if(ci == NULL) semerror(0,"out of memory\n");
+    ci->data = (NCConstant**)ecalloc(sizeof(NCConstant*)*initial);
     ci->alloc = initial;
     ci->length = 0;
     return ci;
@@ -726,45 +596,206 @@ builddatalist(int initial)
 void
 dlappend(Datalist* dl, NCConstant* constant)
 {
-    if(dl->length >= dl->alloc) dlextend(dl);
-    if(constant == NULL) constant = &nullconstant;
-    dl->data[dl->length++] = *constant;
+    if(dl->readonly) abort();
+    if(dl->length >= dl->alloc)
+	dlextend(dl);
+    dl->data[dl->length++] = (constant);
 }
 
-NCConstant
+void
+dlset(Datalist* dl, size_t pos, NCConstant* constant)
+{
+    ASSERT(pos < dl->length);
+    dl->data[pos] = (constant);
+}
+
+NCConstant*
+dlremove(Datalist* dl, size_t pos)
+{
+    int i;
+    NCConstant* con = NULL;
+    ASSERT(dl->length > 0 && pos < dl->length);	
+    con = dl->data[pos];
+    for(i=pos+1;i<dl->length;i++)
+        dl->data[i-1] = dl->data[i];
+    dl->length--;
+    return con;
+}
+
+void
+dlinsert(Datalist* dl, size_t pos, Datalist* insertion)
+{
+    int i;
+    int len1 = datalistlen(dl);
+    int len2 = datalistlen(insertion);
+    int delta = len1 - pos;
+    dlsetalloc(dl,len2+len1+1);
+
+   
+    /* move contents of dl up to make room for insertion */
+    if(delta > 0)
+        memmove(&dl->data[pos+len2],&dl->data[pos],delta*sizeof(NCConstant*));
+    dl->length += len2;
+    for(i=0;i<len2;i++) {
+	NCConstant* con = insertion->data[i];
+	con = cloneconstant(con);
+        dl->data[pos+i] = con;
+    }    
+}
+
+/* Convert a datalist to a compound constant */
+NCConstant*
 builddatasublist(Datalist* dl)
 {
 
-  NCConstant d;
-  d.nctype = NC_COMPOUND;
-  d.lineno = (dl->length > 0?dl->data[0].lineno:0);
-  d.value.compoundv = dl;
-  d.filled = 0;
+  NCConstant* d = nullconst();
+  d->nctype = NC_COMPOUND;
+  d->lineno = (dl->length > 0?dl->data[0]->lineno:0);
+  setconstlist(d,dl);
+  d->filled = 0;
   return d;
 
 }
 
-/*! Function to free an allocated datalist.
+/* Convert a subsequence of a datalist to its own datalist */
+Datalist*
+builddatasubset(Datalist* dl, size_t start, size_t count)
+{
+    Datalist* subset;
 
-  This function is used to free an individual datalist
-  object.  It is possible, hypothetically, that a
-  datalist will appear in the middle of a set of datalists,
-  in which case we'll need to determine that and shuffle around
-  'next' pointers.  For the time being, this assumes that
-  we are freeing a datalist which was allocated locally
-  and must be discarded.
+    if(dl == NULL || start >= datalistlen(dl)) return NULL;
+    if((start + count) > datalistlen(dl))
+        count = (datalistlen(dl) - start);
+    subset = (Datalist*)ecalloc(sizeof(Datalist));
+    subset->readonly = 1;
+    subset->length = count;
+    subset->alloc = count;
+    subset->data = &dl->data[start];
+    return subset;
+}
 
-  Using this function instead of an inline 'free' just in
-  case we ever want to extend it, we won't have to go back
-  and re-write a bunch of stuff. I hope.
+/* Deep copy */
+Datalist*
+clonedatalist(Datalist* dl)
+{
+    int i;
+    size_t len;
+    Datalist* newdl;
 
-  @param dlist Pointer to datalist object being freed.
-
- */
-void dlfree(Datalist **dlist) {
-
-  if(*dlist) free(*dlist);
-  dlist = NULL;
+    if(dl == NULL) return NULL;
+    len = datalistlen(dl);
+    newdl = builddatalist(len);
+    /* initialize */
+    for(i=0;i<len;i++) {
+	NCConstant* con = datalistith(dl,i);
+	con = cloneconstant(con);
+	dlappend(newdl,con);
+    }
+    return newdl;
 }
 
 
+/* recursive helpers */
+
+void
+reclaimconstant(NCConstant* con)
+{
+    if(con == NULL) return;
+    switch (con->nctype) {
+    case NC_STRING:
+	if(con->value.stringv.stringv != NULL)
+	    efree(con->value.stringv.stringv);
+	break;
+    case NC_OPAQUE:
+	if(con->value.opaquev.stringv != NULL)
+	    efree(con->value.opaquev.stringv);
+	break;
+    case NC_COMPOUND:
+#ifdef VERIFY
+	{int pos;
+	if((pos=verify(alldatalists,con->value.compoundv)) >= 0) {
+	    dumpdatalist(listget(alldatalists,pos),"XXX");
+	    abort();
+	}
+	}
+#endif
+        reclaimdatalist(con->value.compoundv);
+	con->value.compoundv = NULL;
+	break;
+    default: break;
+    }
+    efree(con);
+
+}
+
+void
+reclaimdatalist(Datalist* list)
+{
+   int i;
+   if(list == NULL) return;
+   if(!list->readonly) {
+	if(list->data != NULL) {
+	    for(i=0;i<list->length;i++) {
+	        NCConstant* con = list->data[i];
+	        if(con != NULL) reclaimconstant(con);
+            }
+        }
+    }
+    freedatalist(list);
+}
+
+/* Like reclaimdatalist, but do not try to reclaim contained constants */
+void
+freedatalist(Datalist* list)
+{
+   if(list == NULL) return;
+   if(!list->readonly) {
+        efree(list->data);
+        list->data = NULL;
+    }
+    efree(list);
+}
+
+void
+reclaimalldatalists(void)
+{
+    int i;
+    for(i=0;i<listlength(alldatalists);i++) {
+        Datalist* di = listget(alldatalists,i);
+	if(di != NULL)
+	    reclaimdatalist(di);
+    }
+    efree(alldatalists);
+    alldatalists = NULL;    
+}
+
+static void
+flattenR(Datalist* result, Datalist* data, int rank, int depth)
+{
+    int i;
+    NCConstant* con;
+
+    if(rank == depth) return;
+    if(datalistlen(data) == 0) return;
+    for(i=0;i<datalistlen(data);i++) {
+	con = datalistith(data,i);
+        if(depth < rank - 1) {
+	    /* Is this is a char list, then we might have short depth */
+	    if(islistconst(con))
+	        flattenR(result,compoundfor(con),rank,depth+1);
+	    else
+	        dlappend(result,con);
+        } else { /* depth == rank -1, last dimension */
+	    dlappend(result,con);
+	}
+    }
+}
+
+/* Produce a new list that is the concat of all the leaf constants */
+Datalist*
+flatten(Datalist* list,int rank)
+{
+    Datalist* result = builddatalist(0);
+    flattenR(result,list,rank,0);
+    return result;
+}
